/* 
 * CinemaSix Media Player
 * Copyright (c) 2004, Sujal Patel
 * Copyright (C) 2001, Billy Biggs <vektor@dumbterm.net>,
 *                     Håkan Hjort <d95hjort@dtek.chalmers.se>
 * $Id: DVDFile.cpp,v 1.3 2004/07/09 10:10:55 smpatel Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include <cascade/Cascade.h>
#include <cascade/util/CascadeFSObject.h> 
#include <cascade/util/CascadeDirBrowser.h> 
#include <cascade/util/CascadeFile.h> 

#include "dvbpsi/dvbpsi.h"
#include "dvbpsi/psi.h"
#include "dvbpsi/pat.h"
#include "dvbpsi/descriptor.h"
#include "dvbpsi/pmt.h"

#include "DVDFile.h"

#include "config.h"

#if defined(HAVE_INTTYPES_H)
#include <inttypes.h>
#elif defined(HAVE_STDINT_H)
#include <stdint.h>
#endif

#include <dvdread/dvd_reader.h>
#include <dvdread/ifo_types.h>
#include <dvdread/ifo_read.h>

extern int XXXSMP_too_late_at_night_eof_hack;


DVDFile::DVDFile() : File()
{
    m_pFile = NULL;
    m_nNumChapters = 0;
    m_nNumVOBs = 0;
    m_nCurrentSectorCount = 0;
    m_nVTSNumber = 0;
    m_ulCurrentVOBNumber = 0;
}


DVDFile::~DVDFile()
{
    if(m_pFile != NULL)
	delete m_pFile;
}


bool DVDFile::Open(const char *pFileName)
{
    dvd_reader_t *dvd;
    ifo_handle_t *ifo_file;
    tt_srpt_t *tt_srpt;
    int j;

    dvd = DVDOpen(pFileName);
    if( !dvd ) {
        fprintf( stderr, "Couldn't open DVD: %s\n", pFileName);
        return -1;
    }

    ifo_file = ifoOpen( dvd, 0 );
    if( !ifo_file ) {
        fprintf( stderr, "Can't open VMG info.\n" );
        DVDClose( dvd );
        return -1;
    }
    tt_srpt = ifo_file->tt_srpt;

    printf( "There are %d titles.\n", tt_srpt->nr_of_srpts );
    bool nothing_found = true;
    int best_title = 0;
    int best_title_ifo_size = 0;
    int best_vtsnum = 0;
    int best_numchaps = 0;
    int best_ttnnum = 0;

    for (int i = 0; i < tt_srpt->nr_of_srpts; i++)
    {
        int vtsnum, ttnnum, chapts;
	
        vtsnum = tt_srpt->title[i].title_set_nr;
        ttnnum = tt_srpt->title[i].vts_ttn;
        chapts = tt_srpt->title[i].nr_of_ptts;
	
	char name[2048];
	struct stat statbuf;

	sprintf(name, "%s/VTS_%02d_0.IFO", pFileName, vtsnum);
	if (stat(name, &statbuf) == 0)
	{
	    printf( "VTS==%d, TTN=%d, ", vtsnum, ttnnum );
	    printf( "Chapters==%d, ", chapts);
	    printf ("IFO Size %ld\n", statbuf.st_size);
	    if (statbuf.st_size > best_title_ifo_size)
	    {
		best_title_ifo_size = statbuf.st_size;
		best_title = i; 
		best_vtsnum = vtsnum;
		best_numchaps = chapts;
		best_ttnnum = ttnnum;
		nothing_found = false;
	    }
	}
    }
    if (nothing_found)
    {
	DVDClose( dvd );
	return -1;
    }
    printf ("Title #%d (VTS %d) looks good!\n", best_title, best_vtsnum);
    m_nVTSNumber = best_vtsnum;

    ifo_handle_t *vts_file;
    vts_ptt_srpt_t *vts_ptt_srpt;
    pgc_t *cur_pgc;
    int pgcnum;
    bool first = true;

    for (int i = 0; i < 32; i++)
    {
	char name[2048];
	struct stat statbuf;

	sprintf(name, "%s/VTS_%02d_%d.VOB", pFileName, m_nVTSNumber, i);
	if (stat(name, &statbuf) == 0)
	{
	    VOBSectorStart[i] = m_nCurrentSectorCount;
	    m_nCurrentSectorCount += statbuf.st_size / 2048;
	    m_nNumVOBs++;
	    printf ("%s Starts at Sector %x\n", name, VOBSectorStart[i]);
	    first = false;
	}
	else if (first)
	{
	    VOBSectorStart[i] = -1;
	    m_nNumVOBs++;
	    printf ("%s Starts at Sector %x\n", name, VOBSectorStart[i]);
	}
	else
	{
	    break;
	}
    }
    printf ("Ending Sector %x\n", m_nCurrentSectorCount);

    vts_file = ifoOpen( dvd, m_nVTSNumber );
    if( !vts_file ) {
	fprintf( stderr, "Can't open info file for title %d.\n", m_nVTSNumber );
	DVDClose( dvd );
	return -1;
    }

    vts_ptt_srpt = vts_file->vts_ptt_srpt;
    int k = 0;
    for (j = 0; j < best_numchaps; ++j) {
	int start_cell;
	int pgn;

	pgcnum = vts_ptt_srpt->title[ best_ttnnum - 1 ].ptt[ j ].pgcn;
	pgn = vts_ptt_srpt->title[ best_ttnnum - 1 ].ptt[ j ].pgn;
	cur_pgc = vts_file->vts_pgcit->pgci_srp[ pgcnum - 1 ].pgc;
	start_cell = cur_pgc->program_map[ pgn - 1 ] - 1;
	printf("Cell: %3i has VOB ID: %3i, Cell ID: %3i\n", j + 1,
	   cur_pgc->cell_position[j].vob_id_nr, cur_pgc->cell_position[j].cell_nr);

	printf( "\tChapter %3d [PGC %2d, PG %2d] starts at Cell %2d [sector %x-%x]\n",
		j, pgcnum, pgn, start_cell,
		cur_pgc->cell_playback[ start_cell ].first_sector,
		cur_pgc->cell_playback[ start_cell ].last_sector );
	ChapterSectors[k] = cur_pgc->cell_playback[ start_cell ].first_sector;
	k++;
	if (VOBSectorStart[1] != -1) /* VTS_**_1.VOB Exists */
	{
	    if (ChapterSectors[k-1] < VOBSectorStart[1]) /* Chapter starts in Menu VOB */
	    {
		if (cur_pgc->cell_playback[ start_cell ].last_sector < VOBSectorStart[1])
		{
		    k = 0; /* Chapter completely inside the menu, drop all prev. chapters */ 
		}
		else
		{
		    ChapterSectors[k-1] = VOBSectorStart[1]; /* Set new chapter start */
		}
	    }
	}
    }
    m_nNumChapters = best_numchaps;

    ifoClose( ifo_file );
    DVDClose( dvd );
    strcpy(m_Source, pFileName);
    return OpenChapter(0);
}

bool DVDFile::OpenChapter(int nChapter)
{
    char name[2048];
    int nVobNumWeWant = -1;
again:
    printf ("Looking for Sector %x Chapter %d\n", ChapterSectors[nChapter], nChapter);
    for (int i = 0; i < m_nNumVOBs; i++)
    {
	if (((i + 1) < m_nNumVOBs) &&
	    (VOBSectorStart[i] != -1) &&
	    (VOBSectorStart[i] <= ChapterSectors[nChapter]) &&
	    (VOBSectorStart[i+1] > ChapterSectors[nChapter]))
	{
	    if (i != 0)
	    {
		nVobNumWeWant = i;
		break;
	    }
	    else
	    {
		/* Shouldn't need this code anymore, caught above? */
		printf ("Skipping %d because it's in the Menu VOB\n", i);
		nChapter++;
		goto again;
	    }
	}
    }

    Close();

    sprintf(name, "%s/VTS_%02d_%d.VOB", m_Source, m_nVTSNumber, nVobNumWeWant);
    m_pFile = new CascadeFile;
    m_pFile->SetName(name);
    m_OK = m_pFile->Open("r");

    if (m_OK) {
	m_ulCurrentVOBNumber = nVobNumWeWant;
	m_ulCurrentPos = (ChapterSectors[nChapter] -
            VOBSectorStart[nVobNumWeWant]) * 2048;
	m_pFile->SeekForward(m_ulCurrentPos);

	printf ("File Opened == %s, Seeked to sector %x\n", name,
	    (ChapterSectors[nChapter] - VOBSectorStart[nVobNumWeWant]));
	FillBuffer();
    } else {
	printf("Disk file '%s' FAILED to open\n", name);
    }
    return m_OK;
}

void DVDFile::Close()
{
    if (m_pFile != NULL) {
	m_pFile->Close();
	delete m_pFile;
	m_pFile = NULL;
    }
}

bool 
DVDFile::TryNextFile()
{
    int fl = strlen(m_Source);
    for (int i = fl; i > 0; i--) {
        if (isdigit(m_Source[i])) {
                m_Source[i]++;
                if (m_Source[i] == 58)
                {
                    m_Source[i] = '0';
                    if (i > 1) m_Source[i-1]++;
                }
                break;
        }
    }
    Close();
    bool ok = Open(m_Source);
    if (ok) {
        printf ("EOF! Got %s as next file\n", m_Source);
    } else {
        printf ("EOF! There is no next file\n");
    }
    return ok;
}


void
DVDFile::FillBuffer()
{
    /* XXXSMP Use shift ops and inline function? */
    if ((m_bits_consumed / 8) >= (INPUT_BUFFER_SIZE/2)) {
	memmove(m_buffer, m_buffer+(INPUT_BUFFER_SIZE/2),
	    (INPUT_BUFFER_SIZE/2));
	m_buffered_bits -= (INPUT_BUFFER_SIZE/2)*8;
	m_bits_consumed -= (INPUT_BUFFER_SIZE/2)*8;
    }
    while ((!m_eof) && ((m_buffered_bits / 8) <= (INPUT_BUFFER_SIZE/2))) {
	int fd = fileno((FILE *)m_pFile->GetHandle());
	int n = read(fd, m_buffer + (m_buffered_bits / 8), 32768);
	if (n > 0) {
	    m_buffered_bits += n*8;
	    m_ulCurrentPos += n;
	} else {
	    char name[2048];
	    sprintf(name, "%s/VTS_%02d_%d.VOB", m_Source, m_nVTSNumber,
		(++m_ulCurrentVOBNumber)+ 1);
	    printf ("Trying Next File: %s\n", name);
	    delete m_pFile;
	    m_pFile = new CascadeFile;
	    m_pFile->SetName(name);
	    m_OK = m_pFile->Open("r");
	    m_ulCurrentPos = 0;
	    if (!m_OK)
	    {
		printf ("EOF! DVD All done!\n");
		m_eof = 1;
		XXXSMP_too_late_at_night_eof_hack = 1;
		return;
	    }
	}
    }
}

void
DVDFile::Seek(int x)
{
    m_pFile->SeekForward(x);
}


void
DVDFile::SeekB(int x)
{
    m_pFile->SeekBackward(x);
}

bool
DVDFile::ForwardChapter()
{
    int nCurrentSector = m_ulCurrentPos / 2048 + VOBSectorStart[m_ulCurrentVOBNumber];
    int i;

    for (i = 0; i < m_nNumChapters; i++)
    {
	if (nCurrentSector < ChapterSectors[i])
	{
	    break;
	}
    }

    if ((i == nCurrentSector) || (i >= m_nNumChapters) || i < 0)
	return false;

    m_buffered_bits = 0;
    m_bits_consumed = 0;
    m_eof = 0;
    m_OK = 0;
    OpenChapter(i);

    return true;
}

bool
DVDFile::BackwardChapter()
{
    int nCurrentSector = m_ulCurrentPos / 2048 + VOBSectorStart[m_ulCurrentVOBNumber];
    int i;

    for (i = 0; i < m_nNumChapters; i++)
    {
	if (nCurrentSector < ChapterSectors[i])
	{
	    break;
	}
    }

    if (((i - 2) == nCurrentSector) || ((i-2) >= m_nNumChapters) || (i-2) < 0)
	return false;

    m_buffered_bits = 0;
    m_bits_consumed = 0;
    m_eof = 0;
    m_OK = 0;
    OpenChapter(i - 2);

    return true;
}
