/* 
 * CinemaSix Media Player
 * Copyright (c) 2004, Sujal Patel
 * $Id: DiskFile.cpp,v 1.2 2004/06/26 18:38:32 smpatel Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include <cascade/Cascade.h>
#include <cascade/util/CascadeFSObject.h> 
#include <cascade/util/CascadeDirBrowser.h> 
#include <cascade/util/CascadeFile.h> 

#include "dvbpsi/dvbpsi.h"
#include "dvbpsi/psi.h"
#include "dvbpsi/pat.h"
#include "dvbpsi/descriptor.h"
#include "dvbpsi/pmt.h"

#include "DiskFile.h"


extern int XXXSMP_too_late_at_night_eof_hack;


DiskFile::DiskFile() : File()
{
    m_pFile = NULL;
}


DiskFile::~DiskFile()
{
    if(m_pFile != NULL)
	delete m_pFile;
}


bool DiskFile::Open(const char *pFileName)
{
    strcpy(m_Source, pFileName);
    m_pFile = new CascadeFile;
    m_pFile->SetName(pFileName);
    m_OK = m_pFile->Open("r");

    if (m_OK) {
	printf("Disk file '%s' opened\n", pFileName);
	FillBuffer();
    } else {
	printf("Disk file '%s' FAILED to open\n", pFileName);
    }
    return m_OK;
}

void DiskFile::Close()
{
    if(m_pFile != NULL) {
	m_pFile->Close();
	delete m_pFile;
	m_pFile = NULL;
    }
}

bool 
DiskFile::TryNextFile()
{
    int fl = strlen(m_Source);
    for (int i = fl; i > 0; i--) {
        if (isdigit(m_Source[i])) {
                m_Source[i]++;
                if (m_Source[i] == 58)
                {
                    m_Source[i] = '0';
                    if (i > 1) m_Source[i-1]++;
                }
                break;
        }
    }
    Close();
    bool ok = Open(m_Source);
    if (ok) {
        printf ("EOF! Got %s as next file\n", m_Source);
    } else {
        printf ("EOF! There is no next file\n");
    }
    return ok;
}


void
DiskFile::FillBuffer()
{
    /* XXXSMP Use shift ops and inline function? */
    if ((m_bits_consumed / 8) >= (INPUT_BUFFER_SIZE/2)) {
	memmove(m_buffer, m_buffer+(INPUT_BUFFER_SIZE/2),
	    (INPUT_BUFFER_SIZE/2));
	m_buffered_bits -= (INPUT_BUFFER_SIZE/2)*8;
	m_bits_consumed -= (INPUT_BUFFER_SIZE/2)*8;
    }
    while ((!m_eof) && ((m_buffered_bits / 8) <= (INPUT_BUFFER_SIZE/2))) {
	int fd = fileno((FILE *)m_pFile->GetHandle());
	int n = read(fd, m_buffer + (m_buffered_bits / 8), 32768);
	if (n > 0) {
	    m_buffered_bits += n*8;
	} else {
	    int i,fl = strlen(m_Source);
	    for (i = fl; i > 0; i--) {
		if (isdigit(m_Source[i])) {
			m_Source[i]++;
			if (m_Source[i] == 58)
			{
			    m_Source[i] = '0';
			    if (i > 1) m_Source[i-1]++;
			}
			break;
		} else if(m_Source[i] == '/') {
			// Entering directory section - quit looking
			i=0;
			break;
		}
	    }
	    if(i == 0) {
		printf ("EOF! No digits in filename, so we will exit\n");
		m_eof = 1;
		XXXSMP_too_late_at_night_eof_hack = 1;
	    } else {
	    	printf ("EOF! Trying %s as next file\n", m_Source);
	    	delete m_pFile;
	    	m_pFile = new CascadeFile;
	   	m_pFile->SetName(m_Source);
	  	bool ok = m_pFile->Open("r");
	    	if (ok) {
		    // Do Nothing
	    	} else {
		    m_eof = 1;
		    XXXSMP_too_late_at_night_eof_hack = 1;
		}
	    }
	}
    }
}

#define BYTES_PER_ATSC_PACKET 188
void
DiskFile::Seek(int x)
{
  //m_pFile->SeekForward(x);

  int iJump = x / BYTES_PER_ATSC_PACKET;
  int fd = fileno((FILE *)m_pFile->GetHandle());
  __off64_t new_pos = (__off64_t) iJump;
  for (int i = 0; i < BYTES_PER_ATSC_PACKET; i++) {
    lseek64(fd, new_pos, SEEK_CUR);
  }
}


void
DiskFile::SeekB(int x)
{
    Seek(-x);
}
