/* 
 * CinemaSix Media Player
 * Copyright (c) 2004, Sujal Patel
 * Copyright (c) 2004-2005, Andrew Oliver
 * $Id: main.cpp,v 1.21 2005/05/21 19:35:55 aoliver Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>

#include <cascade/Cascade.h>
#include "MPEGDecoderExperimental.h"
#include <cascade/app/CascadeInput.h> 
#include <cascade/util/CascadeFSObject.h> 
#include <cascade/util/CascadeDirBrowser.h> 
#include <cascade/util/CascadeFile.h> 
#include <cascade/widgets/CascadeSpinnerWidget.h> 
#include <cascade/widgets/CascadeButtonWidget.h> 

#include "File.h"
#include "util.h"
#include "InputStreamPS.h"
#include "InputStreamTS.h"
#include "ProgramInfo.h"
#include "mpeg.h"
#include "MyMPEGDecoder.h"
#include "MpegInfoSpinner.h"
#include "KeyboardWindow.h"
#include "main.h"

//#define DEBUG_WINDOW

/* XXXSMP This file requires a lot of cleanup! */


int XXXSMP_too_late_at_night_eof_hack = 0;
int nGlobalFilledBytes = 0;
char g_pCurrentURL[512] = { 'h', 't', 't', 'p', ':', '/', '/', 0 };


class MpegInfoDirWindow : public CascadeDirBrowser
{
protected:
    virtual void OnMaterialize();
};


class MpegInfoWindow : public CascadeWindow
{
public:
    MpegInfoWindow() : m_pKeyboard(0), m_nButtonActive(0), m_bPlayOff(true),
	m_bRecentUp(false), m_ulRecentItemNumber(0) {}
    void ButtonStateChgCB(const CascadeButtonWidget *pButton);
    KeyboardWindow* m_pKeyboard;
protected:
    virtual void OnPaint(CascadeBitmap & bitmap);
    virtual void OnMaterialize();
    virtual bool OnKeyDown(u32 nKey);
    virtual bool OnNavKey(u32 nKey, CascadeWindow*& pProposedNewFocusWnd);
private:
    CascadeButtonWidget m_FileButton;
    CascadeButtonWidget m_HTTPButton;
    CascadeButtonWidget m_RecentButton;
    CascadeButtonWidget m_ExitButton;
    CascadeButtonWidget m_PlayButton;
    int m_nButtonActive;
    bool m_bPlayOff;
    bool m_bRecentUp;
    u32  m_ulRecentItemNumber;
};

class MpegDecoderBustedWindow : public CascadeWindow, CascadeTimer
{
public:
    MpegDecoderBustedWindow() { m_nSec = 10; };
    void ItBroke();
protected:
    virtual void OnPaint(CascadeBitmap & bitmap);
    virtual void OnTimer(u32 nTimerID);
private:
    int m_nSec;
};

void MpegDecoderBustedWindow::OnPaint(CascadeBitmap & bitmap)
{
    CascadeFont font;
    CascadeFont font2;

    CascadeFont::Attributes attr("system", CascadeFont::kNormal, 36);
    CascadeFont::Attributes attr2("system", CascadeFont::kNormal, 24);
    font.SetAttributes(attr);
    font2.SetAttributes(attr2);

    CascadeColor colorBlack(0, 0, 0);
    CascadeColor colorBeige(220, 175, 120);
    CascadeColor colorWhite(255, 255, 255);
    CascadeRect myRect(GetRectAbsolute());

    // paint the window background
    bitmap.FillRect(myRect, colorBlack);

    char string[256];
    CascadeDims textDims(font.GetTextExtent(string));
    CascadeDims textDims2(font2.GetTextExtent(string));
    CascadePoint textPoint(myRect.x + 20, myRect.y + 25);

    CascadeFile imageFile; 
    imageFile.SetName("Sad.png");
    imageFile.Open("r");
    CascadeBitmap ImageBitmap;
    ImageBitmap.CreateFromBitmapFile(imageFile);

    CascadePoint imagePoint(myRect.w/2 - 250/2, myRect.h/2 - 250/2);
    CascadeRect imageRect(0, 0, 250, 250);
    bitmap.Blit(imagePoint, ImageBitmap, imageRect);

    sprintf (string, "An Unrecoverable Error Has Occured!");
    bitmap.TextOut(textPoint, string, font, colorBeige, TEXTOUT_TOP);
    textPoint.y += textDims.h;
    textPoint.y += 15; /* Some Space */

    strcpy(string, "From time to time, while playing video, unrecoverable errors may occur");
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);
    textPoint.y += textDims2.h;

    sprintf(string, "Your HD1000 will reboot in %d seconds", m_nSec);
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);
    textPoint.y += textDims2.h;
}

void MpegDecoderBustedWindow::ItBroke()
{
    SetTimer(1000);
}

void MpegDecoderBustedWindow::OnTimer(u32 nTimerID)
{
    m_nSec--;
    Redraw();
    if (m_nSec == 0)
    {
	system("/sbin/reboot");
    }
}

void PleaseWaitWindow::OnPaint(CascadeBitmap & bitmap)
{
    CascadeRect myRect(GetRectAbsolute());

    CascadeFile imageFile; 
    imageFile.SetName("/usr/local/roku/graphics/common/please-wait.png");
    imageFile.Open("r");
    CascadeBitmap ImageBitmap;
    ImageBitmap.CreateFromBitmapFile(imageFile);

    CascadePoint imagePoint(myRect.w/2 - 314/2, myRect.h/2 - 94/2);
    CascadeRect imageRect(0, 0, 314, 94);
    bitmap.Blit(imagePoint, ImageBitmap, imageRect);
}

void
crashhandler(int code)
{
    printf ("CinemaSix has crashed!\n");
    printf ("Code %d\n", code);
    printf ("Last crash point: %s:%d\n", crash_file, crash_line);
    exit(-1);
}

int main(int argc, const char ** argv)
{
    MPEGPlayerApp app;
    signal(SIGSEGV, crashhandler);
    signal(SIGBUS, crashhandler);

    return app.Run(argc, argv);
}


void
MPEGPlayerWindow::OnPaint(CascadeBitmap & bitmap)
{
    CascadeRect rect(GetRectAbsolute());
    CascadeColor color(0, 0, 0, 0);
    bitmap.FillRect(rect, color);

#ifdef DEBUG_WINDOW
    rect.w /= 2;
    rect.h /= 2;
    rect.x += rect.w / 2;
    rect.y += rect.w / 2;
    CascadeColor colorX(0, 0, 128, 160);
    bitmap.FillRect(rect, colorX);

    CascadeColor colorYellow(250, 253, 0);
    CascadeFont font;

    CascadeFont::Attributes attr("system", CascadeFont::kNormal, 24);
    font.SetAttributes(attr);

    MPEGPlayerApp *pApp = (MPEGPlayerApp *)CascadeApp::GetApp();
    MyMPEGDecoder *pDec = &pApp->m_player;

    char string[80];
    sprintf (string, "Video Size: %dx%d", pDec->m_VideoWidth, pDec->m_VideoHeight);
    CascadeDims textDims(font.GetTextExtent(string));
    CascadePoint textPoint(rect.x + 10, rect.y + 10);
    bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);

    sprintf (string, "Video Aspect Ratio: %s", MPEGAspectTextTable[pDec->m_Aspect]);
    textPoint.y += textDims.h;
    bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);

    sprintf (string, "Frame Rate: %s", MPEGFPSTextTable[pDec->m_FrameRate]);
    textPoint.y += textDims.h;
    bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);

    int ciPrograms = pDec->m_ProgramInfo.GetNumPrograms();
    sprintf(string, "Number of programs: %d", ciPrograms);
    textPoint.y += textDims.h;
    bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);

    for (int i = 0; i < ciPrograms; i++) {
      sprintf(string, "Program #%d pid/audio-pid/video-pid: %x/%x/%x", 
              i+1,
              pDec->m_ProgramInfo.GetProgramPID(i),
              pDec->m_ProgramInfo.GetProgramAudioPID(i),
              pDec->m_ProgramInfo.GetProgramVideoPID(i));
      textPoint.y += textDims.h;
      bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);
      sprintf(string, "Program #%d types: audio/video: %x/%x", 
              i+1,
              pDec->m_ProgramInfo.GetProgramAudioType(i),
              pDec->m_ProgramInfo.GetProgramVideoType(i));
      textPoint.y += textDims.h;
      bitmap.TextOut(textPoint, string, font, colorYellow, TEXTOUT_TOP);
    }
#endif
}


bool
MPEGPlayerWindow::OnKeyDown(u32 nKey)
{
    MPEGPlayerApp * pApp = (MPEGPlayerApp *)CascadeApp::GetApp();

    switch (nKey)
    {
        case CK_EXIT:
        {
            printf ("Got Exit\n");
            pApp->Terminate(0);
            break;
        }
        case CK_PLAYPAUSE:
            if (m_bIsPaused)
            {
                printf ("Resume Streaming\n");
                pApp->m_player.ResumeStreaming();
                m_bIsPaused = false;
            }
            else
            {
                printf ("Pause Streaming\n");
                pApp->m_player.PauseStreaming();
                m_bIsPaused = true;
            }
            break;
        case CK_ROTATE:
        {
	    pApp->m_player.Reinit();

	    break;
	}
        case CK_INFO:
	    pApp->m_pSpinner->Materialize();
	    pApp->m_pSpinner->RedrawScreen();

	    break;
        case CK_FASTFORWARD:
        case 65563:
        {
            if (pApp->m_player.SupportChapters()) {
                pApp->m_player.PauseStreaming();
                pApp->m_player.FlushStream();

                pApp->m_player.ForwardChapter();
    
                pApp->m_player.ResumeStreaming();
                
                pApp->m_player.Reinit();
            }
            break;
        }
        case CK_REWIND:
        case 65564:
        {
            if (pApp->m_player.SupportChapters()) {
                pApp->m_player.PauseStreaming();
                pApp->m_player.FlushStream();

                pApp->m_player.BackwardChapter();
    
                pApp->m_player.ResumeStreaming();

                pApp->m_player.Reinit();
            }
            break;
        }
	case CK_ZOOM_IN:
	{
#if XXXSMP
	    int orig = p->m_CurrentAC3Track;
	    p->m_CurrentAC3Track++;
	    while (p->m_AC3TracksAvailable[p->m_CurrentAC3Track] == 0)
	    {
		    p->m_CurrentAC3Track++;
		    if (p->m_CurrentAC3Track == 8)
			p->m_CurrentAC3Track = 0;
		    if (p->m_CurrentAC3Track == orig)
			break;
            }
	    RedrawScreen();
#endif
	    break;
	}
	case CK_ZOOM_OUT:
	{
	    pApp->m_player.ChangePrograms(pApp->m_player.m_CurrentProgram + 1);
	    RedrawScreen();
	    break;
	}
	case CK_WEST:
		printf ("West hit\n");
	    break;
	case CK_IDLE_HEARTBEAT:
	    break;
        default:
	    printf ("What does %ld mean?\n", nKey);
            break;
    }
    return CascadeWindow::OnKeyDown(nKey);
}


void
MPEGPlayerWindow::OnTimer(u32 nTimerID)
{
  MPEGPlayerApp * pApp = (MPEGPlayerApp *)CascadeApp::GetApp();
#ifdef DEBUG_WINDOW
  Redraw();
#endif

  // If we're not paused, pretend like there was activity so the screen saver won't come on
  if (!m_bIsPaused) {
    CascadeInput::NotifyScreenActivity();
  }

  static int count = 0;
  if (count == 0) {
    KillTimer(m_TimerID);
    m_TimerID = SetTimer(2000);
    pApp->m_player.ResumeStreaming();
  }
  if (XXXSMP_too_late_at_night_eof_hack) {
    // If we have reached the end of the stream/file, exit the app immediately
    CascadeApp::GetApp()->Terminate(0);
  }
  if (count > 2) {
    printf ("OnTimer(): Bitrate=%0.2f Mbps, Total Read=%0.2f MB\n",
            nGlobalFilledBytes / ((count-1)*2)/ 1024.0/1000.0*8.0,
            nGlobalFilledBytes/1024.0/1000.0);
    if (count == 3 && nGlobalFilledBytes == 0)
      {
        MpegDecoderBustedWindow* p = new MpegDecoderBustedWindow;
        CascadeScreen screen;
        p->SetRectAbsolute(screen.GetRect());
        p->Materialize(CascadeWindow::kTopMost);
        p->ItBroke();
      }
  }
  count++;
}


void
MPEGPlayerApp::OnAppInit()
{
    printf ("%s\n", g_pMediaPlayerDesc);
    printf ("%s\n", g_pMediaPlayerCopyright);
    printf ("CinemaSix comes with ABSOLUTELY NO WARRANTY\n");
    printf ("This is free software, and you are welcome to redistribute it\n");
    printf ("under the terms of the GNU General Public License;\n");
    printf ("See COPYING for more information\n");

    CascadeScreen screen;

    if (GetArgc() >= 2) {
	OnAppInit2(GetArgv()[1]);
    } else {

	m_pWindow = new MpegInfoWindow;
	m_pWindow->SetRectAbsolute(screen.GetRect());
	m_pWindow->Materialize();
#if 0
	m_pDirWindow = new MpegInfoDirWindow;
	m_pDirWindow->SetRectAbsolute(screen.GetRect());
	m_pDirWindow->Materialize();
#endif
    }
    m_pSpinner = new MpegInfoSpinner((CascadeWindow *)(&m_window), &m_player);
    m_pSpinner->Setup();
}


void
MPEGPlayerApp::OnAppInit2(const char* path)
{
    // first check for session priority
    u8 nSessionPriority = 128;

    for (int i = 1; i < GetArgc(); i++)
    {
        const char * pArg = GetArgv()[i];
        if (0 == strcmp(pArg, "-sessionPriority"))
        {
            i++;
            if (i >= GetArgc()) nSessionPriority = 0;
            else nSessionPriority = (u8)atoi(GetArgv()[i]);
        }
        else
        if (0 == strcmp(pArg, "-playCount"))
        {
            i++;
            if (i >= GetArgc()) continue;
            else m_nPlayCount = (u32)atoi(GetArgv()[i]);
        }
        if (0 == strcmp(pArg, "-prog"))
        {
            i++;
            if (i >= GetArgc()) continue;
            else m_player.m_CurrentProgram = (u32)atoi(GetArgv()[i]);
        }
    }
    if (nSessionPriority == 0)
    {
        printf("mpegplayer: session priority must be between 1 and 255\n");
        Terminate(0);
        return;
    }

    CascadeScreen screen;
    m_pPleaseWait = new PleaseWaitWindow;
    m_pPleaseWait->SetRectAbsolute(screen.GetRect());
    m_pPleaseWait->Materialize();
    m_pPleaseWait->Redraw();

    char pList[20][512];
    int nEntries = 0;
    mkdir("/etc/CinemaSix", 0755);
    FILE* f = fopen("/etc/CinemaSix/Recents", "r");

    while (f && fgets(pList[nEntries], 500, f) && nEntries < 20)
    {
	char* n = strchr(pList[nEntries], '\n'); 
	char* r = strchr(pList[nEntries], '\r');
	if (n) *n = 0;
	if (r) *r = 0;

	nEntries++;
    }
    if (f)
    {
	fclose(f);
    }
    f = fopen("/etc/CinemaSix/Recents", "w");
    char buffer[512];
    bool m_bNoEnter = false;
    if (realpath(path, buffer))
    {
	for (int i = 0; i < nEntries; i++)
	{
	    if (strcmp(buffer, pList[i]) == 0)
	    {
		m_bNoEnter = true;
	    }
	}
	if (!m_bNoEnter) fprintf(f, "%s\n", buffer);
    }
    else
    {
	for (int i = 0; i < nEntries; i++)
	{
	    if (strcmp(path, pList[i]) == 0)
	    {
		m_bNoEnter = true;
	    }
	}
	if (!m_bNoEnter) fprintf(f, "%s\n", path);
    }
    if (m_bNoEnter == false && nEntries == 20)
    {
	nEntries = 19;
    }
    for (int i = 0; i < nEntries; i++)
    {
	fprintf(f, "%s\n", pList[i]);
    }
    fclose(f);

    strcpy(m_pPath, path);
    TimedStart* ts = new TimedStart;
    ts->SetTimer(150);
}

void
TimedStart::OnTimer(u32 nTimerID)
{
    MPEGPlayerApp * pApp = (MPEGPlayerApp *)CascadeApp::GetApp();

    KillTimer(nTimerID);
    printf ("OnAppInit2()\n");
    pApp->m_player.StartPlayback(pApp->m_pPath);
    pApp->m_window.Materialize();
}

void Callback(const CascadeString& inPath,
    CascadeDirBrowser::CallbackReason inReason, void * pClientData) 
{
    MPEGPlayerApp * pApp = (MPEGPlayerApp *)CascadeApp::GetApp();
    if (inReason == 1)
	CascadeApp::GetApp()->Terminate(0);
    else
    {
	pApp->OnAppInit2((const char *)inPath);
    }
}


void
MpegInfoDirWindow::OnMaterialize()
{
    CascadeFSObject::TypeSelector* pFSSel1 = new CascadeFSObject::TypeSelector;
    CascadeFSObject::TypeSelector* pFSSel2 = new CascadeFSObject::TypeSelector;
    pFSSel1->SetSelectorTypeFilter(CascadeFSObject::Selector::eDirectories);
    pFSSel2->SetSelectorTypeFilter(CascadeFSObject::Selector::eFiles);
    BrowseDirectory("Choose a Video File", "/tmp/Volumes", "/tmp/Volumes", "",
	pFSSel1, pFSSel2, CascadeDirBrowser::eFiles, Callback, 0);
}

char pRecentList[19][256];
int  nEntries = 0;

void
MpegInfoWindow::OnPaint(CascadeBitmap & bitmap)
{
    CascadeFont font;
    CascadeFont font2;

    CascadeFont::Attributes attr("system", CascadeFont::kNormal, 36);
    CascadeFont::Attributes attr2("system", CascadeFont::kNormal, 24);
    font.SetAttributes(attr);
    font2.SetAttributes(attr2);

    CascadeColor colorBlue(0x26, 0x59, 0x86); /* Roku Blue */
    CascadeColor colorBeige(220, 175, 120);
    CascadeColor colorWhite(255, 255, 255);
    CascadeRect myRect(GetRectAbsolute());

    // paint the window background
    bitmap.FillRect(myRect, colorBlue);

    CascadeFile imageFile; 
    imageFile.SetName("CS.png");
    imageFile.Open("r");
    CascadeBitmap ImageBitmap;
    ImageBitmap.CreateFromBitmapFile(imageFile);

    CascadePoint imagePoint(myRect.w/2 - 475/2, myRect.h/2 - 446/2);
    CascadeRect imageRect(0, 0, 475, 446);
    bitmap.Blit(imagePoint, ImageBitmap, imageRect);

    char string[256];
    CascadeDims textDims(font.GetTextExtent(string));
    CascadeDims textDims2(font2.GetTextExtent(string));
    CascadePoint textPoint(myRect.x + 20, myRect.y + 5);

    sprintf (string, "%s", g_pMediaPlayerDesc);
    bitmap.TextOut(textPoint, string, font, colorBeige, TEXTOUT_TOP);
    textPoint.y += textDims.h;

    sprintf (string, "%s", g_pMediaPlayerCopyright);
    bitmap.TextOut(textPoint, string, font, colorBeige, TEXTOUT_TOP);
    textPoint.y += textDims.h;
    textPoint.y += 15; /* Some Space */

    strcpy(string, "This video player plays all MPEG 2 clips.");
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);
    textPoint.y += textDims2.h;

    strcpy(string, "It can also play unencrypted .VOB files from DVDs");
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);

    textPoint.x = 310;

    strcpy(string, "Select a file to play...");
    textPoint.y = 160 + (textDims2.h);
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);

    strcpy(string, "Play a network stream...");
    textPoint.y = 240 + (textDims2.h);
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);

    strcpy(string, "Recently Played...");
    textPoint.y = 320 + (textDims2.h);
    bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);

    if (m_pKeyboard)
    {
	CascadeFont font3;

	CascadeFont::Attributes attr3("system", CascadeFont::kNormal, 20);
	font3.SetAttributes(attr3);
	CascadeDims textDims3(font3.GetTextExtent(string));

	CascadeColor colorYellow(255, 255, 0);

        CascadeRect r;
	r.x = 575 - 3;
	r.y = 380 - 3;
	r.w = 400 + 6;
	r.h = textDims3.h + 6;
#define BKCOLOR  CascadeColor(164, 164, 164)
#define BKCOLOR2 CascadeColor(52, 52, 52)
	bitmap.FillRect(r, CascadeColor(100, 100, 200));
	bitmap.DrawLine(r.TopLeft(), r.TopRight(), BKCOLOR);
	bitmap.DrawLine(r.TopLeft(), r.BottomLeft(), BKCOLOR);
	bitmap.DrawLine(r.TopRight(), r.BottomRight(), BKCOLOR2);
	bitmap.DrawLine(r.BottomLeft(), r.BottomRight(), BKCOLOR2);

	strcpy(string, g_pCurrentURL);
	textPoint.x = 575;
	textPoint.y = 380;
	bitmap.TextOut(textPoint, string, font3, colorYellow, TEXTOUT_TOP);
    }

    if (m_bRecentUp)
    {
	if (nEntries == 0)
	{
	    FILE* f = fopen("/etc/CinemaSix/Recents", "r");

	    while (f && fgets(pRecentList[nEntries], 250, f) && nEntries < 19)
	    {
		char* n = strchr(pRecentList[nEntries], '\n');
		char* r = strchr(pRecentList[nEntries], '\r');
		if (n) *n = 0;
		if (r) *r = 0;
		nEntries++;
	    }
	    if (f)
	    {
		fclose(f);
	    }
	}

	if (m_ulRecentItemNumber > 100)
	{
	    m_ulRecentItemNumber = 0;
	}
	if (m_ulRecentItemNumber >= (unsigned int)nEntries)
	{
	    m_ulRecentItemNumber = nEntries - 1;
	}

	CascadeRect r;
	CascadeFont font3;

	CascadeFont::Attributes attr3("system", CascadeFont::kNormal, 24);
	font3.SetAttributes(attr3);
	CascadeDims textDims3(font3.GetTextExtent(string));

	r.x = 340 - 5;
	r.y = 75 - 5;
	r.w = 650 + 10;
	r.h = textDims3.h * 19 + 10 + textDims2.h;
#define BKCOLOR  CascadeColor(164, 164, 164)
#define BKCOLOR2 CascadeColor(52, 52, 52)
	bitmap.FillRect(r, CascadeColor(25, 25, 85));
	bitmap.DrawLine(r.TopLeft(), r.TopRight(), BKCOLOR);
	bitmap.DrawLine(r.TopLeft(), r.BottomLeft(), BKCOLOR);
	bitmap.DrawLine(r.TopRight(), r.BottomRight(), BKCOLOR2);
	bitmap.DrawLine(r.BottomLeft(), r.BottomRight(), BKCOLOR2);

	/* Seperator Line */
	r.y = 96;
	r.x ++;
	r.w -= 2;
	bitmap.DrawLine(r.TopLeft(), r.TopRight(), CascadeColor(255,255,255));

	strcpy(string, "Recently Played Items");
	textPoint.x = 560;
	textPoint.y = 75;

	bitmap.TextOut(textPoint, string, font2, colorWhite, TEXTOUT_TOP);

	CascadeColor colorYellow(255, 255, 0);

	r.x = 340 - 3;
	r.y = 100 - 3 + m_ulRecentItemNumber*textDims3.h;
	r.w = 650 + 6;
	r.h = textDims3.h + 6;
#define BKCOLOR  CascadeColor(164, 164, 164)
#define BKCOLOR2 CascadeColor(52, 52, 52)
	bitmap.FillRect(r, CascadeColor(100, 100, 200));
	bitmap.DrawLine(r.TopLeft(), r.TopRight(), BKCOLOR);
	bitmap.DrawLine(r.TopLeft(), r.BottomLeft(), BKCOLOR);
	bitmap.DrawLine(r.TopRight(), r.BottomRight(), BKCOLOR2);
	bitmap.DrawLine(r.BottomLeft(), r.BottomRight(), BKCOLOR2);
	textPoint.x = 340;
	textPoint.y = 100;

	for (int i = 0; i < nEntries; i++)
	{
	    strcpy(string, pRecentList[i]);
	    char* mount = strstr(string, "/tmp/Volumes/");
	    if (mount)
	    {
		memcpy(mount, "[Share]  ", 9);
	    }
	    bitmap.TextOut(textPoint, string, font3, colorYellow, TEXTOUT_TOP);
	    textPoint.y += textDims3.h;
	}
    }
}

void KeyboardEntryCB(const CascadeString & text, void * pClientData)
{
    MpegInfoWindow* miw = (MpegInfoWindow*)pClientData;

    if (text == "Back")
    {
	if (strlen(g_pCurrentURL))
	{
	    g_pCurrentURL[strlen(g_pCurrentURL)-1] = 0;
	    miw->Redraw();
	}
    }
    else if (text == "Done")
    {
	if (g_pCurrentURL[strlen(g_pCurrentURL)-1] != '/')
	{
	    strcat(g_pCurrentURL, (const char*)"/");
	}
	//miw->m_pKeyboard->Vanish();
	//delete miw->m_pKeyboard;
	miw->m_pKeyboard = NULL;
	MPEGPlayerApp * pApp = (MPEGPlayerApp *)CascadeApp::GetApp();
	pApp->OnAppInit2((const char *)g_pCurrentURL);
    }
    else if (text == "Space")
    {
	strcat(g_pCurrentURL, (const char*)" ");
	miw->Redraw();
    }
    else
    {
	strcat(g_pCurrentURL, (const char*)text);
	miw->Redraw();
    }
}

void ButtonStateChangeCB(const CascadeButtonWidget *pButton,
	const CascadeButtonWidget::State &newState, void *pClientData) 
{
    MpegInfoWindow* pMIW = (MpegInfoWindow*)pClientData;

    pMIW->ButtonStateChgCB(pButton);
}

void MpegInfoWindow::ButtonStateChgCB(const CascadeButtonWidget *pButton)
{
    MPEGPlayerApp *pApp = (MPEGPlayerApp *)CascadeApp::GetApp();

    if (pButton == &m_FileButton) {
        CascadeScreen screen;

	pApp->m_pDirWindow = new MpegInfoDirWindow;
	pApp->m_pDirWindow->SetRectAbsolute(screen.GetRect());
	pApp->m_pDirWindow->Materialize();

	m_PlayButton.SetDisabled(false, false);
    }
    if (pButton == &m_HTTPButton) {
	u32 flags = KEYBOARD_NUMBERS | KEYBOARD_SHIFT | KEYBOARD_BACK |
	    KEYBOARD_SPACE | KEYBOARD_DONE | KEYBOARD_SHIFT_NUMBERS;
	m_pKeyboard = new KeyboardWindow(flags);
	if (m_pKeyboard)
	{
	    CascadeDims dims = m_pKeyboard->GetDims();
	    CascadeRect keyboardRect(600, 200, dims.w, dims.h);
	    AddChild(m_pKeyboard);
	    m_pKeyboard->SetRectRelative(keyboardRect);
	    m_pKeyboard->RegisterOnEntryCallback(KeyboardEntryCB, this);
	    m_pKeyboard->Create();
	    Redraw();
	    m_pKeyboard->Activate();
	}
    }
    if (pButton == &m_RecentButton) {
	m_bRecentUp = true;
	m_FileButton.SetDisabled(true, false);
	m_HTTPButton.SetDisabled(true, false);
	m_RecentButton.SetDisabled(true, false);
	m_ExitButton.SetDisabled(true, false);
	SetFocus();
	Redraw();
    }

    if (pButton == &m_ExitButton) {
	pApp->Terminate(0);
    }
}

void
MpegInfoWindow::OnMaterialize()
{
    CascadeFile normalFile; 
    normalFile.SetName(
            "/usr/local/roku/graphics/common/textbutton-narrow-normal.png");
    normalFile.Open("r");
    CascadeBitmap* pNormalBitmap = new CascadeBitmap;
    pNormalBitmap->CreateFromBitmapFile(normalFile);

    CascadeFile highlightFile;
    highlightFile.SetName(
            "/usr/local/roku/graphics/common/textbutton-narrow-highlight.png");
    highlightFile.Open("r");
    CascadeBitmap* pHighlightBitmap = new CascadeBitmap;
    pHighlightBitmap->CreateFromBitmapFile(highlightFile);

    CascadeFile disabledFile;
    disabledFile.SetName(
            "/usr/local/roku/graphics/common/textbutton-narrow-disabled.png");
    disabledFile.Open("r");
    CascadeBitmap* pDisabledBitmap = new CascadeBitmap;
    pDisabledBitmap->CreateFromBitmapFile(disabledFile);

    AddChild(&m_FileButton);
    m_FileButton.SetRectAbsolute(CascadeRect(20, 160, 215, 63));
    m_FileButton.SetText("Play File");
    m_FileButton.SetFocus();
    m_FileButton.Materialize();
    m_FileButton.SetBitmapUp(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_FileButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_FileButton.SetBitmapUp(pHighlightBitmap, CascadeButtonWidget::kActiveFocus);
    m_FileButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveFocus);
    m_FileButton.SetBitmapUp(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_FileButton.SetBitmapDown(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_FileButton.RegisterStateChangeCallback(ButtonStateChangeCB, (void *)this);

    AddChild(&m_HTTPButton);
    m_HTTPButton.SetRectAbsolute(CascadeRect(20, 240, 215, 63));
    m_HTTPButton.SetText("Play URL");
    m_HTTPButton.Materialize();
    m_HTTPButton.SetBitmapUp(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_HTTPButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_HTTPButton.SetBitmapUp(pHighlightBitmap, CascadeButtonWidget::kActiveFocus);
    m_HTTPButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveFocus);
    m_HTTPButton.SetBitmapUp(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_HTTPButton.SetBitmapDown(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_HTTPButton.RegisterStateChangeCallback(ButtonStateChangeCB, (void *)this);

    AddChild(&m_RecentButton);
    m_RecentButton.SetRectAbsolute(CascadeRect(20, 320, 215, 63));
    m_RecentButton.SetText("Play Recent Items");
    m_RecentButton.Materialize();
    m_RecentButton.SetBitmapUp(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_RecentButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_RecentButton.SetBitmapUp(pHighlightBitmap, CascadeButtonWidget::kActiveFocus);
    m_RecentButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveFocus);
    m_RecentButton.SetBitmapUp(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_RecentButton.SetBitmapDown(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_RecentButton.RegisterStateChangeCallback(ButtonStateChangeCB, (void *)this);

    AddChild(&m_ExitButton);
    m_ExitButton.SetRectAbsolute(CascadeRect(60, 450, 215, 63));
    m_ExitButton.SetText("Exit");
    m_ExitButton.Materialize();
    m_ExitButton.SetBitmapUp(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_ExitButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_ExitButton.SetBitmapUp(pHighlightBitmap, CascadeButtonWidget::kActiveFocus);
    m_ExitButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveFocus);
    m_ExitButton.SetBitmapUp(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_ExitButton.SetBitmapDown(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_ExitButton.RegisterStateChangeCallback(ButtonStateChangeCB, (void *)this);

#if 0
    AddChild(&m_PlayButton);
    m_PlayButton.SetRectAbsolute(CascadeRect(500, 450, 215, 63));
    m_PlayButton.SetText("Play");
    m_PlayButton.SetDisabled(true, false);
    m_PlayButton.Materialize();
    m_PlayButton.SetBitmapUp(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_PlayButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveNormal);
    m_PlayButton.SetBitmapUp(pHighlightBitmap, CascadeButtonWidget::kActiveFocus);
    m_PlayButton.SetBitmapDown(pNormalBitmap, CascadeButtonWidget::kActiveFocus);
    m_PlayButton.SetBitmapUp(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_PlayButton.SetBitmapDown(pDisabledBitmap, CascadeButtonWidget::kDisabled);
    m_PlayButton.RegisterStateChangeCallback(ButtonStateChangeCB, (void *)this);
#endif
}

/*
 * m_nButtonActive Map
 *
 * File(0)
 * HTTP(1)
 * Recent(2)
 *
 * Exit(3)    Play(4)
 */

bool
MpegInfoWindow::OnNavKey(u32 nKey, CascadeWindow*& pProposedNewFocusWnd)
{
    if (m_pKeyboard)
    {
	return m_pKeyboard->OnNavKey(nKey, pProposedNewFocusWnd);
    }
    if (m_bRecentUp)
    {
	switch (nKey)
	{
	    case CK_NORTH:
	    {
		m_ulRecentItemNumber--;
		Redraw();
		break;
	    }
	    case CK_SOUTH:
	    {
		m_ulRecentItemNumber++;
		Redraw();
		break;
	    }
	}
	return true;
    }
    switch (nKey)
    {
        case CK_NORTH:
	{
	    switch (m_nButtonActive) {
	        case 0: m_nButtonActive = 0; break;
	        case 1: m_nButtonActive = 0; break;
	        case 2: m_nButtonActive = 1; break;
	        case 3: m_nButtonActive = 2; break;
	        case 4: m_nButtonActive = 2; break;
	    }
	    break;
	}
        case CK_SOUTH:
	{
	    switch (m_nButtonActive) {
	        case 0: m_nButtonActive = 1; break;
	        case 1: m_nButtonActive = 2; break;
	        case 2: m_nButtonActive = 3; break;
	        case 3: m_nButtonActive = 3; break;
	        case 4: m_nButtonActive = 4; break;
	    }
	    break;
	}
        case CK_EAST:
	{
	    switch (m_nButtonActive) {
	        case 0: m_nButtonActive = m_bPlayOff ? 0 : 4; break;
	        case 1: m_nButtonActive = m_bPlayOff ? 1 : 4; break;
	        case 2: m_nButtonActive = m_bPlayOff ? 2 : 4; break;
	        case 3: m_nButtonActive = m_bPlayOff ? 3 : 4; break;
	        case 4: m_nButtonActive = m_bPlayOff ? 4 : 4; break;
	    }
	    break;
	}
        case CK_WEST:
	{
	    switch (m_nButtonActive) {
	        case 0: m_nButtonActive = 0; break;
	        case 1: m_nButtonActive = 1; break;
	        case 2: m_nButtonActive = 2; break;
	        case 3: m_nButtonActive = 3; break;
	        case 4: m_nButtonActive = 3; break;
	    }
	    break;
	}
    }

    switch (m_nButtonActive) {
	case 0: pProposedNewFocusWnd = &m_FileButton; break;
	case 1: pProposedNewFocusWnd = &m_HTTPButton; break;
	case 2: pProposedNewFocusWnd = &m_RecentButton; break;
	case 3: pProposedNewFocusWnd = &m_ExitButton; break;
	case 4: pProposedNewFocusWnd = &m_PlayButton; break;
    }
    return true;
}

bool
MpegInfoWindow::OnKeyDown(u32 nKey)
{
    MPEGPlayerApp *pApp = (MPEGPlayerApp *)CascadeApp::GetApp();

    switch (nKey)
    {
        case CK_EXIT:
            pApp->Terminate(0);
            break;
        case CK_SELECT:
	    if (m_bRecentUp)
	    {
		pApp->OnAppInit2((const char *)pRecentList[m_ulRecentItemNumber]);
		break;
	    }		
        default:
	    printf ("What does %ld mean?\n", nKey);
            break;
    }
    return CascadeWindow::OnKeyDown(nKey);
}
